VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "TEMPerGold"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'==========
'TEMPerGold
'==========
'
'Version: 2.3 August 2012
'Author: Bob Riemersma
'
'This is a VB6 class that reads data from PCsensor's USB HID Thermometer
'product "TEMPer Gold."
'
'
'TEMPer Products
'---------------

'There are number of products in the TEMPer series.  Some have a relative
'humidity sensor.  Some have an external temperature sensor.  Some have
'multiple sensors (internal and remote).
'
'Some early TEMPer products present themselves as a USB Serial (COMx:)
'port.
'
'This version was only written to work with the TEMPer Gold product.  A
'TEMPer Gold has one temperature sensor and a sealed cast metal case with
'a label on it that looks like a thermometer on a white background.  There
'is also one LED that should be on (red) when in operation - though my
'sample never had the LED come on in any circumstances (perhaps mine is a
'dummy, since it really serves no purpose).
'
'Some of these have a gold-tone finish on the USB connector.  Mine does
'not.  Perhaps this is a sign of a mass-market version with a working LED?
'
'
'Hardware Details
'----------------
'
'The sensor used is a Fairchild FM75, which is an enhanced version of the
'LM75.  The FM75 can provide 9 to 12 bit readings where the LM75 was
'limited to 9 bits.  It uses a band-gap sensor, not a thermistor or micro
'thermocouple, etc. as has been stated so wildy in the Linux and Delphi
'communities.
'
'You might find the datasheets on both the LM75 and FM75 interesting.
'
'It appears that the TEMPer Gold sets the FM75 for 12-bit readings when
'you do an open on the HID device.
'
'The interface PCsensor used between the PC and the FM75 I2C device seems
'to be a USB microcontroller from a company named Sonix.
'
'The "calibration" setting seems to be stored in NVRAM of some sort, most
'likely on board the microcontroller.  I'd avoid setting this value very
'often since there might be a limit on write cycles it can accept before
'failing.  Don't be overly concerned about this, just avoid doing it
'constantly.
'
'
'General Notes
'-------------
'
'I realize a lot of what I said above seems tentative and conditional.
'But the fact is the manufacturer offers strikingly little information
'about the device: basically nothing.
'
'The FM75 supports alarm/trigger on a preset temperature and the TEMPer
'Gold seems to support it.  However (a.) there's that lack of info
'getting in the way again, and (b.) that feature was meant for use in
'simple hardware/microcontroller monitors.  We can do the same thing in
'software so we don't really need this functionality anyway.
'
'
'Usage
'-----
'
'   Properties
'   ----------
'
'   Count
'       Read only, Long.
'
'       After Initialize() has been called it reflects the number of
'       TEMPer Gold HID devices found
'
'   Error
'       Read only, String.
'
'       Will be empty ("") on most True returns from method calls.
'       Exception: Calling CloseTEMPer() on a device already closed
'       returns True but sets Error to "Alerady closed."
'
'   IsOpen(Dev)
'       Read only, Boolean.  Dev is a TEMPer device index (0 to n).
'
'       Returns True if the device is open.  False if not, and also
'       sets Error to "Not open" message.
'
'   Methods
'   -------
'
'   CloseTEMPer(Dev)
'       Returns Boolean, True on success.
'
'       Closes specified TEMPer Gold device.
'
'   CloseTEMPers()
'       Returns Boolean, True on success for all devices.
'
'       Calls CloseTEMPer(Dev) for each device.
'
'   CtoF(Celsius)
'       Returns Single.
'
'       Celsius value is converted to Fahrenheit.  This is merely a
'       "conveneience" function for the caller.
'
'   FtoC(Fahrenheit)
'       Returns Single.
'
'       Fahrenheit value is converted to Celsius.  This is merely a
'       "conveneince" function for the caller.
'
'   GetCalibrationCelsius(Dev, Calibration)
'       Returns a Boolean, True on success.
'
'       Dev is a ByVal Long device index and Calibration is a ByRef
'       Integer used to return the current calibration value, -8C
'       to +7C.
'
'   GetCelsius(Dev, Celsius)
'       Returns Boolean, True on success.
'
'       Retrieves the temperature for Dev index (ByVal Long) and
'       the ByRef Single Celsius is used to return the meansurement.
'
'   GetFahrenheit(Dev, Fahrenheit)
'       Returns Boolean, True on success.
'
'       This is a "convenience" wrapper for a call to GetCelsius()
'       and a call to FtoC().
'
'   GetVersion(Dev, Version)
'       Returns Boolean, True on success.
'
'       Retrieves the version info for device Dev (ByVal Long) and
'       returns it via Version (ByRef String).
'
'   Initialize()
'       Returns Boolean, True on success.
'
'       THIS IS AN IMPORTANT CALL.  You must call Initialize()
'       *once* after a new instance of TEMPerGold has been created.
'       It scans for TEMPer Gold devices and sets the Count property.
'
'       If you need to rescan device later, create a New instance of
'       TEMPerGold and then call Initialize() on it.
'
'   OpenTEMPer(Dev)
'       Returns Boolean, True on success.
'
'       Opens communication for data transfer with the device indexed
'       by Dev.
'
'   SetCalibrationCelsius(Dev, Calibration)
'       Returns Boolean, True on success.
'
'       Sets the device Dev's (ByVal Long) calibration offset to
'       Calibration (ByVal Integer).  Valid values are -8C to +7C.
'
'
'Normally you would create an instance of TEMPerGold, call Initialize()
'on it, and based on Count you would call OpenTEMPer() for one or more
'devices.
'
'From there you might call GetVersion() or GetCalibration() or even
'SetCalibration - though I suggest you avoid chaning this extremely
'frequently.  The value you set persists, and the small NVRAM on the
'microcontroller in each device may have a limited number of write
'cycles.
'
'At this point you can call GetCelsius() and/or GetFahrenheit() to take
'readings.  However readings much closer than a second or two probably
'won't show much change in most cases since the metal case of the
'TEMPer Gold imposes a bit of latency anyway.
'
'You can call CloseTEMPer() on a device when finished with it.  From
'the thin docmentation available, this may put the TEMPer Gold device
'into an even lower power state.  You can always call OpenTEMPer() on
'the device later and begin taking more readings.
'
'When done with all of the TEMPer Gold devices you can call the
'CloseTEMPers() method to close out all open devices.  In any case
'when TEMPerGold terminates it closes all open devices itself.
'
'
'Revision History
'----------------
'
'    Date     Version  Change
'===========  =======  ===================================================
'2012-Jul-23  00.0000  o First experiments using the HID Raw Input API in
'                        Windows.
'
'2012-Jul-24  00.0001  o Further experimentation.
'
'2012-Jul-26  00.0002  o Abandoned Raw Input API, trying to reverse
'                        engineer calls to RDingUSB.dll.  First tests that
'                        call this library.
'
'2012-Jul-27  00.0003  o Disassembled RDingUSB.dll and decompiled the
'                        PCsensor .Net sample application for analysis.
'                      o Refined calls to RDingUSB.dll based on this new
'                        information.
'
'2012-Jul-30  00.0004  o First decent results but some holes remain.  Will
'                        need to do more research to refine the calls.
'
'2012-Jul-31  01.0000  o Made a few adjustments such as adding some dummy
'                        read operations to help flip the state of the
'                        device.  First stable, completely functioning
'                        version at last.
'                      o Cleaned up the CrLf produced by calls to
'                        FormatMessage().
'
'2012-Aug-01  01.0001  o Testing revealed a few minor refinments but this
'                        was mostly a change to improve the comments in
'                        the code.
'
'2012-Aug-02  02.0000  o Using information from the disassembly of the
'                        RDingUSB.dll (really mostly just the list of its
'                        imports and exports), began a new version that
'                        will be independent of RDingUSB.dll.
'             02.0001  o Introduced a new method Initialize() and a new
'                        property Count.  With these it will be possible
'                        to support multiple TEMPer Gold units connected
'                        to one machine.
'
'2012-Aug-03  02.0002  o Completely eliminated use of the Raw Input API,
'                        even for enumerating the USB devices.  Now using
'                        a combination of the SetupAPI and Hid API
'                        libraries.  This is clunkier since we have to do
'                        a special open on each device to determine what
'                        it is, but it is more reliable - and what the
'                        SonixUSB.dll/RDingUSB.dll was doing itself.
'
'2012-Aug-03  02.0003  o General optimization to eliminate redundant code
'                        prior to publishing Version 2.
'                      o Minor rewrite to make it easier to potentially
'                        add support for other TEMPer HID products later.
'                      o Streamlined error processing to save more code.
'                      o Added documentation on "Usage" of the TEMPerGold
'                        class.
'

Private Const TEMPerGoldVID As Integer = &HC45
Private Const TEMPerGoldPID As Integer = &H7401
Private Const TEMPerGoldUsagePage As Integer = &HFF00
Private Const TEMPerGoldUsage As Integer = &H1

Private Const INVALID_HANDLE_VALUE As Long = -1&
Private Const OPEN_EXISTING As Long = 3
Private Const API_NULL As Long = 0
Private Const MAX_PATH As Long = 260

Private Enum FILE_ACCESS
    GENERIC_WRITE = &H40000000
    GENERIC_READ = &H80000000
End Enum

Private Enum FILE_SHARING
    FILE_SHARE_READ = &H1&
    FILE_SHARE_WRITE = &H2&
End Enum

Private Enum DIGCF_FLAGS
    DIGCF_PRESENT = &H2
    DIGCF_DEVICEINTERFACE = &H10
End Enum

Private Type GUID
    Data1 As Long
    Data2 As Integer
    Data3 As Integer
    Data4(0 To 7) As Byte
End Type

Private Type HIDP_CAPS
    'We need to use a trick here to avoid VB6's DWord alignment:
    'USAGE  Usage
    'USAGE  UsagePage
    'USHORT  InputReportByteLength
    'USHORT  OutputReportByteLength
    'USHORT  FeatureReportByteLength
    'USHORT  Reserved[17]
    'USHORT  NumberLinkCollectionNodes
    'USHORT  NumberInputButtonCaps
    'USHORT  NumberInputValueCaps
    'USHORT  NumberInputDataIndices
    'USHORT  NumberOutputButtonCaps
    'USHORT  NumberOutputValueCaps
    'USHORT  NumberOutputDataIndices
    'USHORT  NumberFeatureButtonCaps
    'USHORT  NumberFeatureValueCaps
    'USHORT  NumberFeatureDataIndices
    Caps(32) As Integer
End Type

Private Enum HidPCaps
    'Indexes we care about into Caps() above.
    CapUsage
    CapUsagePage
End Enum

Private Type HIDD_ATTRIBUTES
    Size As Long
    'We need to use a trick here to avoid VB6's DWord alignment:
    'USHORT VendorID
    'USHORT ProductID
    'USHORT VersionNumber
    Attribs(2) As Integer
    'Be sure to use Len(HIDD_ATTRIBUTES) and not LenB(HIDD_ATTRIBUTES)!
End Type

Private Enum HidDAttrs
    'Indexes we care about into Attribs() above.
    AttrVendorID
    AttrProductID
End Enum

Private Enum HIDP_STATUSES
    HIDP_STATUS_SUCCESS = &H110000
    HIDP_STATUS_INVALID_PREPARSED_DATA = &HC0110001
End Enum

Private Type SP_DEVICE_INTERFACE_DATA
    cbSize As Long
    InterfaceClassGuid As GUID
    Flags As Long
    Reserved As Long
End Type

Private Type SP_DEVICE_INTERFACE_DETAIL_DATA
    cbSize As Long
    DevicePath(511) As Byte
End Type

Private Declare Function CloseHandle Lib "kernel32" ( _
    ByVal hObject As Long) As Long

Private Declare Function CreateFile Lib "kernel32" Alias "CreateFileW" ( _
    ByVal lpFileName As Long, _
    ByVal dwDesiredAccess As FILE_ACCESS, _
    ByVal dwShareMode As FILE_SHARING, _
    ByVal lpSecurityAttributes As Long, _
    ByVal dwCreationDisposition As Long, _
    ByVal dwFlagsAndAttributes As Long, _
    ByVal hTemplateFile As Long) As Long

Private Declare Function HidD_FreePreparsedData Lib "hid" ( _
    ByVal pPreparsedData As Long) As Long

Private Declare Function FormatMessage Lib "kernel32" Alias "FormatMessageA" ( _
    ByVal dwFlags As Long, _
    ByRef Source As Long, _
    ByVal dwMessageId As Long, _
    ByVal dwLanguageId As Long, _
    ByVal Buffer As String, _
    ByVal nSize As Long, _
    ByRef Arguments As Any) As Long

Private Declare Function HidD_GetAttributes Lib "hid" ( _
    ByVal HidDeviceObject As Long, _
    ByRef Attributes As HIDD_ATTRIBUTES) As Long

Private Declare Function HidP_GetCaps Lib "hid" ( _
    ByVal pPreparsedData As Long, _
    ByRef Capabilities As HIDP_CAPS) As HIDP_STATUSES

Private Declare Function HidD_GetPreparsedData Lib "hid" ( _
    ByVal HidDeviceObject As Long, _
    ByRef pPreparsedData As Long) As Long

Private Declare Sub HidD_GetHidGuid Lib "hid" (ByRef HidGuid As GUID)

Private Declare Function ReadFileAPI Lib "kernel32" Alias "ReadFile" ( _
    ByVal hFile As Long, _
    ByRef Buffer As Any, _
    ByVal nNumberOfBytesToRead As Long, _
    ByRef NumberOfBytesRead As Long, _
    ByVal lpOverlapped As Long) As Long

Private Declare Function SetupDiDestroyDeviceInfoList Lib "setupapi" ( _
    ByVal DeviceInfoSet As Long) As Long

Private Declare Function SetupDiEnumDeviceInterfaces Lib "setupapi" ( _
    ByVal DeviceInfoSet As Long, _
    ByVal DeviceInfoData As Long, _
    ByRef InterfaceClassGuid As GUID, _
    ByVal MemberIndex As Long, _
    ByRef DeviceInterfaceData As SP_DEVICE_INTERFACE_DATA) As Long

Private Declare Function SetupDiGetClassDevs Lib "setupapi" Alias "SetupDiGetClassDevsA" ( _
    ByRef ClassGuid As GUID, _
    ByVal Enumerator As String, _
    ByVal hWndParent As Long, _
    ByVal Flags As DIGCF_FLAGS) As Long

Private Declare Function SetupDiGetDeviceInterfaceDetail Lib "setupapi" _
    Alias "SetupDiGetDeviceInterfaceDetailA" ( _
    ByVal DeviceInfoSet As Long, _
    ByRef DeviceInterfaceData As SP_DEVICE_INTERFACE_DATA, _
    ByVal lpDeviceInterfaceDetailData As Long, _
    ByVal DeviceInterfaceDetailDataSize As Long, _
    ByRef RequiredSize As Long, _
    ByVal lpDeviceInfoData As Long) As Long

Private Declare Function WriteFileAPI Lib "kernel32" Alias "WriteFile" ( _
    ByVal hFile As Long, _
    ByRef Buffer As Any, _
    ByVal nNumberOfBytesToWrite As Long, _
    ByRef NumberOfBytesWritten As Long, _
    ByVal lpOverlapped As Long) As Long

'====== Local Data =====================================================

Private TEMPerNames() As String
Private hTEMPers() As Long
Private HidGuid As GUID
Private CommandGetCalibration() As Byte
Private CommandGetVersion() As Byte
Private CommandReadTemper() As Byte
Private CommandSetCalibration() As Byte
Private CommandSetValue() As Byte
Private mError As String
Private mCount As Long

'====== Public Properties ===============================================

Public Property Get Count() As Long
    Count = mCount
End Property

Public Property Get Error() As String
    Error = mError
End Property

Public Property Get IsOpen(ByVal Dev As Long) As Boolean
    If hTEMPers(Dev) <> INVALID_HANDLE_VALUE Then
        IsOpen = True
    Else
        mError = "Device " & CStr(Dev) & " not open."
    End If
End Property

'====== Public Methods ==================================================

Public Function CloseTEMPer(ByVal Dev As Long) As Boolean
    'Returns True on success (including when already closed).
    
    If DevValid(Dev) Then
        If IsOpen(Dev) Then
            If CloseHid(hTEMPers(Dev)) Then
                CloseTEMPer = True
            Else
                mError = "Device " & CStr(Dev) & " " & mError
            End If
        Else
            CloseTEMPer = True
        End If
    End If
End Function

Public Function CloseTEMPers() As Boolean
    'Returns True if all closes succeeded.
    Dim I As Long
    
    CloseTEMPers = True
    For I = 0 To mCount - 1
        If Not CloseTEMPer(I) Then
            CloseTEMPers = False
        End If
    Next
End Function

Public Function CtoF(ByVal Celsius As Single) As Single
    CtoF = Celsius * (9# / 5#) + 32#
End Function

Public Function FtoC(ByVal Fahrenheit As Single) As Single
    'Included more for symmatry's sake than a real need.  However
    'it could be used in calculating Celsius Calibration values
    'from Fahrenheit values.  You'll want to be careful to round
    'or truncate those explicitly though, since the Calibration
    'parameter is a small Integer value in whole Celsius degrees.
    
    FtoC = (Fahrenheit - 32#) * (5# / 9#)
End Function

Public Function GetCalibrationCelsius(ByVal Dev As Long, ByRef Calibration As Integer) As Boolean
    'Returns True on success.
    Dim Success As Boolean
    Dim Buffer(8) As Byte
    
    If DevValid(Dev) Then
        If IsOpen(Dev) Then
            Success = WriteFile(hTEMPers(Dev), CommandGetCalibration)
            If Success Then
                Success = ReadFile(hTEMPers(Dev), Buffer)
                If Success Then
                    Calibration = CInt(Buffer(3)) \ 16
                    If Calibration And &H8 Then Calibration = Calibration Or &HFFF0
                    GetCalibrationCelsius = True
                End If
            End If
            If Not Success Then mError = "Device " & CStr(Dev) & " " & mError
        End If
    End If
End Function

Public Function GetCelsius(ByVal Dev As Long, ByRef Celsius As Single) As Boolean
    'Returns True on Success.
    '
    'Here we assume the TEMPer Gold's (TEMPerF1.2's) Fairchild FM75
    '(an enhanced LM75) has been set by the TEMPer to use 12-bit
    'temperature resolution instead of its default, poweron, 9-bit
    'resolution matching that of the LM75.
    Dim Success As Boolean
    Dim Buffer(8) As Byte
    Dim Reading As Integer
    
    If DevValid(Dev) Then
        If IsOpen(Dev) Then
            Success = WriteFile(hTEMPers(Dev), CommandReadTemper)
            If Success Then
                Success = ReadFile(hTEMPers(Dev), Buffer)
                If Success Then
                    Reading = CInt(Buffer(3)) * 16 Or CInt(Buffer(4)) \ 16
                    If Reading And &H800 Then Reading = Reading Or &HF000
                    Celsius = CSng(Reading) * 0.0625
                    GetCelsius = True
                End If
            End If
            If Not Success Then mError = "Device " & CStr(Dev) & " " & mError
        End If
    End If
End Function

Public Function GetFahrenheit(ByVal Dev As Long, ByRef Fahrenheit As Single) As Boolean
    Dim Celsius As Single
    
    If GetCelsius(Dev, Celsius) Then
        Fahrenheit = CtoF(Celsius)
        GetFahrenheit = True
    End If
End Function

Public Function GetVersion(ByVal Dev As Long, ByRef Version As String) As Boolean
    'Returns True on success.
    Dim Success As Boolean
    Dim Buffer(8) As Byte
    Dim Part1 As String
    Dim Part2 As String
    Dim FindEnd As Long

    If DevValid(Dev) Then
        If IsOpen(Dev) Then
            Success = WriteFile(hTEMPers(Dev), CommandGetVersion)
            If Success Then
                'Here is a terrible hack, but it seems to be necessary.  We
                'must do the reads 8 bytes of data at a time, but the version
                'info is a bit longer than 8 characters so we need to do a
                'second read.  However we don't know the actual length to
                'expect, nor do we get a terminating NUL or anything to go
                'by.
                '
                'We're taking the first 8, and then after taking the second
                '8 characters we backscan comparing the two until a non-match
                'which we hope tells us where the end is.  This is quite a
                'troubling procedure:
                '
                '   o What if we have 7, exactly 8, or 17 characters?
                '   o What if we stumble over a duplicated character that
                '     is supposed to be the same in both parts?
                '
                'Ugly this may be, but I see others have followed the same
                'path, probably leading back to PCsensor's own .Net example
                'program!
                Success = ReadFile(hTEMPers(Dev), Buffer)
                If Success Then
                    Part1 = StrConv(MidB$(Buffer, 2), vbUnicode)
                    Success = ReadFile(hTEMPers(Dev), Buffer)
                    If Success Then
                        Part2 = StrConv(MidB$(Buffer, 2), vbUnicode)
                        For FindEnd = 8 To 1 Step -1
                            If Mid$(Part2, FindEnd, 1) <> Mid$(Part1, FindEnd, 1) Then Exit For
                        Next
                        Version = Part1 & Left$(Part2, FindEnd)
                        GetVersion = True
                    End If
                End If
            End If
            If Not Success Then mError = "Device " & CStr(Dev) & " " & mError
        End If
    End If
End Function

Public Function Initialize() As Boolean
    Dim Result As Long
    Dim I As Long
    
    If mCount = 0 Then
        Result = EnumerateHids(mCount, TEMPerNames)
        If Result Then
            'Note special case here, see EnumerateHids():
            If Result <> -1 Then mError = DecodeAPIErrors(Result)
        Else
            If mCount = 0 Then
                mError = "No TEMPer devices."
            Else
                ReDim hTEMPers(mCount - 1)
                For I = 0 To mCount - 1
                    hTEMPers(I) = INVALID_HANDLE_VALUE
                Next
                Initialize = True
            End If
        End If
    Else
        mError = "Already initialized."
    End If
End Function

Public Function OpenTEMPer(ByVal Dev As Long) As Boolean
    'Returns True on success.
    
    If DevValid(Dev) Then
        If IsOpen(Dev) Then
            mError = "Already open."
        Else
            If OpenHid(hTEMPers(Dev), GENERIC_READ Or GENERIC_WRITE, TEMPerNames(Dev)) Then
                OpenTEMPer = True
            Else
                mError = "Device " & CStr(Dev) & " " & mError
            End If
        End If
    End If
End Function

Public Function SetCalibrationCelsius(ByVal Dev As Long, ByVal Calibration As Integer) As Boolean
    'Returns True on success.
    Dim Success As Boolean
    Dim Buffer() As Byte 'Dynamic to permit array assignment below.
    
    If DevValid(Dev) Then
        If IsOpen(Dev) Then
            If -8 > Calibration Or Calibration > 7 Then
                mError = "Calibration value must be within the range -8C to +7C."
            Else
                Success = WriteFile(hTEMPers(Dev), CommandSetCalibration)
                If Success Then
                    'Dummy read here.  Not sure why but it makes the next write "take."
                    ReDim Buffer(8)
                    ReadFile hTEMPers(Dev), Buffer
                    
                    Buffer = CommandSetValue 'Assign initial bytes.
                    Buffer(5) = (Calibration And &HF) * 16
                    Success = WriteFile(hTEMPers(Dev), Buffer)
                    If Success Then
                        'Dummy read here.  Not sure why but it makes a subsequent "getcal" work.
                        ReadFile hTEMPers(Dev), Buffer
                        
                        SetCalibrationCelsius = True
                    End If
                End If
                If Not Success Then mError = "Device " & CStr(Dev) & " " & mError
            End If
        End If
    End If
End Function

'====== Private Methods =================================================

Private Function CloseHid(ByRef Handle As Long) As Boolean
    If CloseHandle(Handle) Then
        Handle = INVALID_HANDLE_VALUE
        mError = ""
        CloseHid = True
    Else
        mError = DecodeAPIErrors(Err.LastDllError)
    End If
End Function

Private Function DecodeAPIErrors(ByVal ErrorCode As Long) As String
    Const FORMAT_MESSAGE_FROM_SYSTEM As Long = &H1000&
    Dim Msg As String
    Dim MsgLen As Long

    If ErrorCode = -1 Then
        DecodeAPIErrors = "Buffer too short."
    Else
        Msg = Space$(MAX_PATH)
        MsgLen = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, _
                               ByVal 0&, _
                               ErrorCode, _
                               0, _
                               Msg, _
                               Len(Msg), _
                               ByVal 0&)
        If MsgLen > 0 Then
            DecodeAPIErrors = Left(Msg, MsgLen)
            MsgLen = InStr(DecodeAPIErrors, vbCr)
            If MsgLen Then DecodeAPIErrors = Left$(DecodeAPIErrors, MsgLen - 1)
        Else
            DecodeAPIErrors = "Unknown Error."
        End If
    End If
End Function

Private Function DevValid(ByVal Dev As Long) As Boolean
    DevValid = 0 <= Dev And Dev < mCount
    If Not DevValid Then mError = "Device " & CStr(Dev) & " is out of range."
End Function

Private Function EnumerateHids(ByRef DevCount As Long, ByRef DevNames() As String) As Long
    'Returns 0 on success, non-0 error code on failure.  Error code -1 is
    'an OpenHidAtts() or CloseHid() failure and mError has already been set!
    Dim hDevInfo As Long
    Dim EnumSuccess As Boolean
    Dim I As Long
    Dim OpenCloseSuccess As Boolean
    Dim MiscSuccess As Boolean
    Dim DevIntfData As SP_DEVICE_INTERFACE_DATA
    Dim DevIntfDetailData As SP_DEVICE_INTERFACE_DETAIL_DATA
    Dim RequiredSize As Long
    Dim DevName As String
    Dim hDev As Long
    Dim pPreparsedData As Long
    Dim HidPCaps As HIDP_CAPS
    Dim HidDAttributes As HIDD_ATTRIBUTES
    
    DevCount = 0
    hDevInfo = SetupDiGetClassDevs(HidGuid, _
                                   vbNullString, _
                                   0, _
                                   DIGCF_PRESENT Or DIGCF_DEVICEINTERFACE)
    If hDevInfo <> INVALID_HANDLE_VALUE Then
        Do
            DevIntfData.cbSize = Len(DevIntfData)
            EnumSuccess = SetupDiEnumDeviceInterfaces(hDevInfo, _
                                                      0, _
                                                      HidGuid, _
                                                      I, _
                                                      DevIntfData)
            If EnumSuccess Then
                'Call with null pointer to get RequiredSize:
                SetupDiGetDeviceInterfaceDetail hDevInfo, _
                                                DevIntfData, _
                                                0, _
                                                0, _
                                                RequiredSize, _
                                                0
                'Oddity here: We must set cbSize to everything prior to
                '.DevicePath (cbSize), plus 1 byte for the terminating NUL.
                DevIntfDetailData.cbSize = Len(DevIntfDetailData.cbSize) + 1
                MiscSuccess = SetupDiGetDeviceInterfaceDetail(hDevInfo, _
                                                              DevIntfData, _
                                                              VarPtr(DevIntfDetailData), _
                                                              RequiredSize, _
                                                              ByVal 0&, _
                                                              0)
                If MiscSuccess Then
                    'RequiredSize is the entire structure, which includes the other
                    'structure members (cbSize) and a terminating NUL (1).
                    DevName = _
                        StrConv(LeftB$(DevIntfDetailData.DevicePath, _
                                       RequiredSize - Len(DevIntfDetailData.cbSize) - 1), _
                                vbUnicode)
                    OpenCloseSuccess = OpenHid(hDev, 0, DevName)
                    If OpenCloseSuccess Then
                        MiscSuccess = HidD_GetPreparsedData(hDev, pPreparsedData)
                        If MiscSuccess Then
                            MiscSuccess = HidP_GetCaps(pPreparsedData, HidPCaps) = HIDP_STATUS_SUCCESS
                            If MiscSuccess Then
                                With HidPCaps
                                    If .Caps(CapUsagePage) = TEMPerGoldUsagePage And _
                                       .Caps(CapUsage) = TEMPerGoldUsage Then
                                        With HidDAttributes
                                            .Size = Len(HidDAttributes)
                                            MiscSuccess = HidD_GetAttributes(hDev, HidDAttributes)
                                            If MiscSuccess Then
                                                If .Attribs(AttrVendorID) = TEMPerGoldVID And _
                                                   .Attribs(AttrProductID) = TEMPerGoldPID Then
                                                    ReDim Preserve DevNames(DevCount)
                                                    DevNames(DevCount) = DevName
                                                    DevCount = DevCount + 1
                                                End If
                                            End If
                                        End With
                                    End If
                                End With
                            End If
                            HidD_FreePreparsedData pPreparsedData
                        End If
                        OpenCloseSuccess = CloseHid(hDev)
                        If Not OpenCloseSuccess Then EnumerateHids = -1 'Special return value, see comments above.
                    Else
                        EnumerateHids = -1 'Special return value, see comments above.
                    End If
                End If
                If Not MiscSuccess Then EnumerateHids = Err.LastDllError
                If Not (OpenCloseSuccess And MiscSuccess) Then Exit Do
            End If
            I = I + 1
        Loop While EnumSuccess
        
        SetupDiDestroyDeviceInfoList hDevInfo
    Else
        EnumerateHids = Err.LastDllError
    End If
End Function

Private Function MakeCommand(ParamArray ByteVals() As Variant) As Byte()
    Dim Command() As Byte
    Dim I As Long
    
    ReDim Command(UBound(ByteVals))
    For I = 0 To UBound(ByteVals)
        Command(I) = ByteVals(I) And &HFF
    Next
    MakeCommand = Command
End Function

Private Function OpenHid( _
    ByRef Handle As Long, _
    ByVal Access As FILE_ACCESS, _
    ByVal FileName As String) As Boolean
    'Returns True on success.
    
    Handle = CreateFile(StrPtr(FileName), _
                        Access, _
                        FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                        0, _
                        OPEN_EXISTING, _
                        0, _
                        0)
    If Handle = INVALID_HANDLE_VALUE Then
        mError = DecodeAPIErrors(Err.LastDllError)
    Else
        OpenHid = True
        mError = ""
    End If
End Function

Private Function ReadFile(ByVal Handle As Long, ByRef Buffer() As Byte) As Boolean
    'Returns True on success, sets mError.
    '
    'Assumes LBound(Buffer) = 0 and Ubound(Buffer) tells us how much to read.
    Dim BytesRead As Long
    
    If ReadFileAPI(Handle, Buffer(0), UBound(Buffer) + 1, BytesRead, 0) Then
        ReadFile = True
        mError = ""
    Else
        mError = DecodeAPIErrors(Err.LastDllError)
    End If
End Function

Private Function WriteFile(ByVal Handle As Long, ByRef Buffer() As Byte) As Boolean
    'Returns True on success, sets mError.
    '
    'Assumes LBound(Buffer) = 0 and Ubound(Buffer) tells us how much to write.
    Dim BytesWritten As Long
    
    If WriteFileAPI(Handle, Buffer(0), UBound(Buffer) + 1, BytesWritten, 0) Then
        WriteFile = True
        mError = ""
    Else
        mError = DecodeAPIErrors(Err.LastDllError)
    End If
End Function

'====== Event Handlers ==================================================

Private Sub Class_Initialize()
    HidD_GetHidGuid HidGuid
    CommandGetCalibration = MakeCommand(0, 1, &H82, &H77, 1, 0, 0, 0, 0)
    CommandGetVersion = MakeCommand(0, 1, &H86, &HFF, 1, 0, 0, 0, 0)
    CommandReadTemper = MakeCommand(0, 1, &H80, &H33, 1, 0, 0, 0, 0)
    CommandSetCalibration = MakeCommand(0, 1, &H85, &HDD, 1, 1, 0, 0, 0)
    CommandSetValue = MakeCommand(0, 1, &H81, &H55, 1, 0, 0, 0, 0)
    mError = "Not open."
End Sub

Private Sub Class_Terminate()
    CloseTEMPers
End Sub
